# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import errno
import http.server
import io
import ipaddress
import itertools
import json
import logging
import re
import socket
import socketserver
import urllib

from . import errors
from . import stats
from . import version

log = logging.getLogger("http")

# Common HTTP status codes
# See https://tools.ietf.org/html/rfc2616#section-6.1.1
CONTINUE = 100
OK = 200
NO_CONTENT = 204
PARTIAL_CONTENT = 206
BAD_REQUEST = 400
FORBIDDEN = 403
NOT_FOUND = 404
METHOD_NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
CONFLICT = 409
REQUEST_URI_TOO_LARGE = 414
REQUESTED_RANGE_NOT_SATISFIABLE = 416
INTERNAL_SERVER_ERROR = 500

# Taken from asyncore.py. Treat these as expected error when reading or writing
# to client connection.
_DISCONNECTED = frozenset((
    errno.ECONNRESET,
    errno.ENOTCONN,
    errno.ESHUTDOWN,
    errno.ECONNABORTED,
    errno.EPIPE
))

# Sentinel for lazy initialization, ensuring that we initialize only once.
_UNKNOWN = object()


class Error(Exception):

    def __init__(self, code, message, content_range=None):
        self.code = code
        self.message = message
        self.content_range = content_range

    def __str__(self):
        return str(self.message)


class Server(socketserver.ThreadingMixIn,
             http.server.HTTPServer):
    """
    Threaded HTTP server.
    """
    daemon_threads = True

    # A callable called for every request.
    app = None

    # Clock used to profile connections. The default NullClock does not
    # profile anything to minimize overhead. Set to stats.Clock to enable
    # profiling.
    clock_class = stats.NullClock

    def __init__(self, server_address, RequestHandlerClass, prefer_ipv4=False):
        super().__init__(
            server_address, RequestHandlerClass, bind_and_activate=False)

        # Close old socket created in parent constructor.
        if self.socket:
            self.socket.close()

        try:
            self.create_socket(prefer_ipv4)
            self.server_bind()
            self.server_activate()
        except BaseException:
            self.server_close()
            raise

    def create_socket(self, prefer_ipv4=False):
        """
        Create socket with correct socket family.
        If prefer_ipv4 is set, and both ipv4 and ipv6 addresses are
        available, pick the first ipv4 address.
        """
        host, port = self.server_address

        addresses = list(find_addresses(host, port=port))

        # If IPv4 is preferred, sort addresses according to address family, so
        # that IPv4 addresses are before IPv6 addresses. Addresses is a list of
        # tuples, address family being the first item in the tuple.
        log.debug("Prefer IPv4: %s", prefer_ipv4)
        if prefer_ipv4:
            addresses.sort(key=lambda x: x[0] != socket.AF_INET)

        log.debug("Available network interfaces: %s", addresses)
        self.address_family = addresses[0][0]

        # Create new socket with correct address family.
        log.debug(
            "Creating server socket with family=%s and type=%s",
            self.address_family,
            self.socket_type)
        try:
            self.socket = socket.socket(self.address_family, self.socket_type)
        except OSError as e:
            raise errors.ServerStartupError(
                "Failed to create socket for address={!r} address family={} "
                "socket type={} prefer IPv4={}. Underlying error is {}".format(
                    host,
                    self.address_family,
                    self.socket_type,
                    prefer_ipv4,
                    e))

    def server_bind(self):
        """
        Override server_bind to make server_address uniform.
        """
        super().server_bind()

        # TCPServer.server_bind() overwrites server_address with
        # socket.getsockname(), which is a tuple of two values when bound to
        # IPv4 interface (e.g. ('0.0.0.0', 54322)) and tuple of four values
        # when bound to IPv6 interface (e.g. ('::', 54322, 0, 0)). To make
        # server address uniform, ensure it's always a tuple of two values -
        # hostname/IP address and port number.
        self.server_address = self.server_address[:2]


class Connection(http.server.BaseHTTPRequestHandler):
    """
    HTTP server connection.

    The server create a new instance for every connection, and then
    process all requests on this connection using the server.app
    callable.
    """

    protocol_version = "HTTP/1.1"

    # Avoids possible delays when sending very small response.
    disable_nagle_algorithm = True

    # The maximum length of the request line:
    # https://tools.ietf.org/html/rfc2616#section-5.1. This limit the size of
    # the request URI. The spec does not define this length, and common
    # browsers support up to 65536 bytes. For imageio purposes, we don't need
    # to support long URIs, so we use small value.
    max_request_line = 4096

    # Number of second to wait for recv() or send() on unauthorized
    # connections.  When the timeout expires we close the connection.
    # Authorized connections get a larger timeout using the ticket
    # inactivity timeout.
    timeout = 15

    # For generating connection ids. Start from 1 to match the connection
    # thread name.
    _counter = itertools.count(1)

    def setup(self):
        self.id = next(self._counter)
        log.info("OPEN connection=%s client=%s",
                 self.id, self.address_string())
        super().setup()
        # Per connection context, used by application to cache state.
        self.context = Context()
        self.clock = self.server.clock_class()
        self.clock.start("connection")

    def finish(self):
        self.clock.stop("connection")
        self.context.close()
        del self.context
        log.info("CLOSE connection=%s client=%s %s",
                 self.id, self.address_string(), self.clock)
        try:
            super().finish()
        except socket.error as e:
            if e.args[0] not in _DISCONNECTED:
                raise
            log.debug("Client disconnected client=%s", self.address_string())

    def handle_one_request(self):
        """
        Override to dispatch requests to server.app and improve error
        handling and logging.

        See the original version here:
        https://github.com/python/cpython/blob/2.7/Lib/BaseHTTPServer.py
        https://github.com/python/cpython/blob/master/Lib/http/server.py
        """
        try:
            self.raw_requestline = self.rfile.readline(
                self.max_request_line + 1)
            if len(self.raw_requestline) > self.max_request_line:
                self.requestline = ''
                self.request_version = ''
                self.command = ''
                self.send_error(REQUEST_URI_TOO_LARGE)
                return

            if not self.raw_requestline:
                log.debug("Empty request line, client disconnected")
                self.close_connection = 1
                return

            if not self.parse_request():
                return

            self.server.app(Request(self), Response(self))
            self.wfile.flush()

        except socket.timeout as e:
            log.warning("Timeout reading or writing to socket: %s", e)
            self.close_connection = 1
        except socket.error as e:
            if e.args[0] not in _DISCONNECTED:
                raise
            log.debug("Client disconnected: %s", e)
            self.close_connection = 1

    def address_string(self):
        """
        Override to avoid slow and unneeded name lookup.
        """
        return self.client_address[0]

    def log_request(self, msg):
        """
        Override to disable request logs from send_response().
        """

    def log_error(self, fmt, *args):
        """
        Override to log errors from send_error() to our log instead of
        stderr. Since these are always client errors, log a warning.
        """
        log.warning(fmt, *args)

    def connection_error(self):
        """
        Return the error number from the underlying socket, or 0 if the socket
        has no error.
        """
        return self.connection.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)

    def version_string(self):
        """
        Used in Server header.
        """
        return "imageio/" + version.string

    def set_timeout(self, timeout):
        log.debug("Setting connection timeout to %s seconds", timeout)
        self.connection.settimeout(timeout)


class Request:

    def __init__(self, con):
        self._con = con
        self._uri = con.path
        if "?" in con.path:
            path, self._query_string = con.path.split("?", 1)
        else:
            path, self._query_string = con.path, ""
        self._path = urllib.parse.unquote(path)
        self._query = _UNKNOWN
        self._content_length = _UNKNOWN
        self._length = _UNKNOWN
        self._range = _UNKNOWN
        self._content_range = _UNKNOWN

    @property
    def context(self):
        """
        Return connection context dict. Can be used to cache per
        connection that can be used by future request on this
        connection.

        Objects added to this dict are discarded when the connection is
        closed. If the objects implement "close", they are also closed.
        """
        return self._con.context

    @property
    def connection_id(self):
        return self._con.id

    @property
    def clock(self):
        """
        Return connection clock.

        Request handlers may use this clock to time operations. Timed
        operations are logged when closing a connection.
        """
        return self._con.clock

    @property
    def headers(self):
        return self._con.headers

    @property
    def version(self):
        return self._con.request_version

    @property
    def method(self):
        return self._con.command

    @property
    def uri(self):
        """
        Return raw request URI including path and query string. This is mostly
        useful for logging.
        """
        return self._uri

    @property
    def path(self):
        """
        Return unquoted path component from the request URI.
        """
        return self._path

    @property
    def query(self):
        """
        Return parsed query string dict.

        If key appears multiple times, the last key=value pair wins:

            key=1&key=2&key -> {"key": ""}

        Both keys and values are decoded to unicode on python 2 and str on
        python 3. Value that cannot be decoded to utf-8 are dropped silently.
        """
        if self._query is _UNKNOWN:
            self._query = dict(urllib.parse.parse_qsl(
                self._query_string, keep_blank_values=True))
        return self._query

    @property
    def content_length(self):
        """
        Return parsed Content-Length header value, or None if the header is
        missing.

        Raises:
            http.Error if Content-Length is not a positive integer.
        """
        if self._content_length is _UNKNOWN:
            value = self._con.headers.get("content-length")
            if value is not None:
                # Requiring valid value here avoid this error handling code
                # from application code.
                # Note: set to None on errors to avoid failures when handling
                # the error.
                try:
                    value = int(value)
                except ValueError:
                    self._content_length = None
                    raise Error(BAD_REQUEST,
                                "Invalid Content-Length: {!r}".format(value))
                if value < 0:
                    self._content_length = None
                    raise Error(BAD_REQUEST,
                                "Negative Content-Length: {!r}".format(value))
            self._content_length = value
        return self._content_length

    @property
    def length(self):
        """
        Return the number of unread bytes left in the request body, or None if
        the request does not have valid content length.
        """
        if self._length is _UNKNOWN:
            self._length = self.content_length
        return self._length

    @property
    def range(self):
        if self._range is _UNKNOWN:
            value = self.headers.get("range")
            if value is not None:
                self._range = Range.parse(value)
            else:
                self._range = None
        return self._range

    @property
    def content_range(self):
        if self._content_range is _UNKNOWN:
            value = self.headers.get("content-range")
            if value is not None:
                self._content_range = ContentRange.parse(value)
            else:
                self._content_range = None
        return self._content_range

    @property
    def client_addr(self):
        return self._con.address_string()

    def read(self, n=None):
        # TODO: support chunked encoding.
        if not self.length:
            return b""

        if n is None or n > self._length:
            n = self._length

        data = self._con.rfile.read(n)
        self._length -= len(data)

        return data

    def readinto(self, buf):
        # TODO: support chunked encoding.
        if not self.length:
            return 0

        length = min(len(buf), self._length)

        with memoryview(buf)[:length] as view:
            n = self._con.rfile.readinto(view)

        self._length -= n
        return n

    def connection_lost(self):
        """
        Return True if the underlying socket was disconnected.
        """
        return self._con.connection_error() in _DISCONNECTED

    def set_connection_timeout(self, timeout):
        self._con.set_timeout(timeout)


class Response:

    def __init__(self, con):
        self._con = con
        self.status_code = OK
        self.headers = Headers({"content-length": 0})
        self._started = False

    @property
    def started(self):
        """
        Return True if the response started.
        """
        return self._started

    def send_info(self, status_code, message=None):
        """
        Send informational status response (1xx) to the client, before
        sending the actual response.

        This can be called multiple times before sending the actual
        response.
        """
        if self._started:
            raise AssertionError("Response already sent")

        msg = self._con.responses[status_code][0].encode("ascii")
        self._con.wfile.write(b"HTTP/1.1 %d %s\r\n\r\n" % (status_code, msg))
        self._con.wfile.flush()

    def send_error(self, e):
        """
        Send an error response for error e.
        """
        if self._started:
            raise AssertionError("Response already sent")

        self.status_code = e.code

        # Adding newline makes it easier to debug from the command line.
        body = str(e).encode("utf-8") + b"\n"

        # Content-type header of the response is set to text/plain to mark
        # content as non-interpretable by the client and thus avoid false
        # positive from security analysis tools when we send requested URL
        # back as part of, say, 404 response
        self.headers["content-type"] = "text/plain; charset=UTF-8"

        self.headers["content-length"] = len(body)
        if e.content_range is not None:
            self.headers["content-range"] = e.content_range
        self.write(body)

    def send_json(self, obj):
        """
        Send a JSON response.
        """
        if self._started:
            raise AssertionError("Response already sent")

        self.status_code = OK
        body = json.dumps(obj).encode("utf-8") + b"\n"
        self.headers["content-length"] = len(body)
        self.headers["content-type"] = "application/json"
        self.write(body)

    def close_connection(self):
        """
        Mark the connection for closing when the request completes.
        """
        self.headers["connection"] = "close"
        self._con.close_connection = True

    def write(self, data):
        """
        Write data to the response body.

        NOTE: This method is replaced with the underlying file write method on
        the first call. Do not try to cache this method!
        """
        self._started = True

        b = io.BytesIO()
        self._write_header(b)

        # For small payload, it is faster to copy the data and write in one
        # syscall.
        # TODO: 4096 is a guess, measure what is the best value.
        if len(data) < 4096:
            b.write(data)
            self._con.wfile.write(b.getvalue())
        else:
            self._con.wfile.write(b.getvalue())
            self._con.wfile.write(data)

        # This avoids name lookup on the next calls to write.
        self.write = self._con.wfile.write

    def _write_header(self, b):
        """
        Write HTTP header to buffer b, avoiding one syscall per line in python
        2.7.
        """
        if self.status_code in self._con.responses:
            msg = self._con.responses[self.status_code][0].encode("latin1")
        else:
            msg = b""

        # Write response line.
        b.write(b"%s %d %s\r\n" % (
                self._con.protocol_version.encode("latin1"),
                self.status_code,
                msg))

        # Write default headers.
        b.write(b"server: %s\r\n" %
                self._con.version_string().encode("latin1"))
        b.write(b"date: %s\r\n" %
                self._con.date_time_string().encode("latin1"))

        # Write user headers.
        for name, value in self.headers.items():
            # Encoding entire line to allow using integer value, for example
            # content-length.
            # Note: content-disposition may contain unicode values, so we must
            # encode headers using utf-8.
            b.write(("%s: %s\r\n" % (name, value)).encode("utf-8"))

        # End header.
        b.write(b"\r\n")


class Context(dict):
    """
    A dict with close interface, closing all closable values.
    """

    def close(self):
        for v in self.values():
            if hasattr(v, "close"):
                try:
                    v.close()
                except Exception:
                    log.exception("Error closing %s", v)


class Headers(dict):
    """
    Dictionarry optimized for keeping HTTP headers.
    """

    def __setitem__(self, name, value):
        """
        Override to enforce lowercase keys. This make it easier to get
        and use values from response headers when processing the
        response.
        """
        dict.__setitem__(self, name.lower(), value)


class Range:
    """
    HTTP Range header.

    See https://tools.ietf.org/html/rfc7233#section-2.1 for details.
    """

    _rx = re.compile(r"bytes=(\d*)-(\d*)$")

    def __init__(self, first, last):
        self.first = first
        self.last = last

    @classmethod
    def parse(cls, header):
        """
        Parse Range header.

        Based on webob.byterange.Range, with several modifications to
        suite our our use case:

        - Raises on invalid or unsupported range so server code does not
          need to do error checking. This also ensure that invalid range will
          not be considered as non-range request which can lead to image
          corruption when the server return the wrong data to the client.

        - Matches better the HTTP spec, using "first" and "last" instead of
          "start" and "stop". The last byte position is kept as is, and not
          incremented by one.

        - Reject invalid range without first or last byte positions, negative
          first, or junk at the end.

        - Does not support case insensitive unit or extra space, since the HTTP
          spec does not specify them.

        - Does not accept None header, only string. We don't want to
          help sloppy programmers.

        Like webob, multiple ranges are not supported, but unlike webob, this
        will reject the entire request, instead of using only the first range.

        Raise:
            http.Error(REQUESTED_RANGE_NOT_SATISFIABLE) if the range is
            invalid, or contain multiple ranges.
        """
        m = cls._rx.match(header)
        if not m:
            raise Error(
                REQUESTED_RANGE_NOT_SATISFIABLE,
                "Cannot satisfy range {!r}, invalid range or multiple ranges"
                .format(header))

        first, last = m.groups()
        if not first:
            if not last:
                # "bytes=-"
                raise Error(
                    REQUESTED_RANGE_NOT_SATISFIABLE,
                    "Cannot satisfy range {!r}, no first or last"
                    .format(header))
            # "bytes=-99"
            return cls(-int(last), None)

        first = int(first)
        if not last:
            # "bytes=0-"
            return cls(first, None)

        last = int(last)
        if first > last:
            raise Error(
                REQUESTED_RANGE_NOT_SATISFIABLE,
                "Cannot satisfy range {!r}, first > last"
                .format(header))

        # "bytes=0-99"
        return cls(first, last)


class ContentRange:
    """
    HTTP ContentRange header.

    See https://tools.ietf.org/html/rfc7233#section-4.2 for details.
    """

    _rx = re.compile(r"bytes (\d+)-(\d+|\*)/(\d+|\*)$")

    def __init__(self, first, last, complete):
        self.first = first
        self.last = last
        self.complete = complete

    @classmethod
    def parse(cls, header):
        """
        Parse ContentRange header.

        This is similar to webob.byterange.ContentRange, with this
        differences:

        - Raises on invalid content range, so server code does not need
          to repeat the error checking. This also ensures that invalid
          header will not be ignored, which will write data to the wrong
          position in an image, corrupting the image.

        - Matches better HTTP spec, using "first", "last", and
          "complete", instead of "start", "stop", and "length", which is
          confusing with content-length.

        - Does not implement the unsatisfied-range format (*/complete)
          which we don't support in a PUT request.

        - Does not accept None, only string. We don't want to help
          sloppy programmers.

        Raise:
            http.Error(BAD_REQUEST) if the range is invalid.
        """
        m = cls._rx.match(header)
        if not m:
            raise Error(
                BAD_REQUEST,
                "Invalid Content-Range {!r}".format(header))

        first, last, complete = m.groups()

        first = int(first)

        if last == "*":
            last = None
        else:
            last = int(last)
            if last < first:
                raise Error(
                    BAD_REQUEST,
                    "Invalid Content-Range {!r}, first > last"
                    .format(header))

        if complete == "*":
            complete = None
        else:
            complete = int(complete)
            if last is not None and last >= complete:
                raise Error(
                    BAD_REQUEST,
                    "Invalid Content-Range {!r}, last >= complete"
                    .format(header))

        return cls(first, last, complete)


class Router:
    """
    Route requests to registered requests handlers.
    """

    ALLOWED_METHODS = frozenset(
        ['GET', 'PUT', 'PATCH', 'POST', 'DELETE', 'OPTIONS', 'HEAD'])

    def __init__(self, routes):
        self._routes = [(re.compile(pattern), handler)
                        for pattern, handler in routes]

    def __call__(self, req, resp):
        with req.clock.run("dispatch"):
            try:
                self.dispatch(req, resp)
                # If request has no content, we need to invoke write()
                # to send the headers. Doing it here simplify client
                # code that want to return the default 200 OK response.
                if not resp.started:
                    resp.write(b"")
            except Exception as e:
                if isinstance(e, socket.error) and req.connection_lost():
                    log.debug("Client disconnected: %s", e)
                    resp.close_connection()
                elif resp.started:
                    # Already started the response, close the connection.
                    log.exception("Request aborted after starting response")
                    resp.close_connection()
                else:
                    # Don't expose internal errors to client.
                    if not isinstance(e, Error):
                        e = Error(
                            INTERNAL_SERVER_ERROR,
                            "Server failed to perform the request, check logs")
                    # Log the error.
                    if e.code >= INTERNAL_SERVER_ERROR:
                        log.exception("Server error")
                    elif e.code >= BAD_REQUEST:
                        log.error("Client error: %s", e)
                    # Did we read the entire request content?
                    if req.length and req.length > 0:
                        resp.close_connection()
                    resp.send_error(e)

    def dispatch(self, req, resp):
        if req.method not in self.ALLOWED_METHODS:
            raise Error(METHOD_NOT_ALLOWED,
                        "Invalid method {!r}".format(req.method))

        path = req.path
        for route, handler in self._routes:
            match = route.match(path)
            if match:
                try:
                    method = getattr(handler, req.method.lower())
                except AttributeError:
                    raise Error(METHOD_NOT_ALLOWED,
                                "Method {!r} not defined for {!r}"
                                .format(req.method, path))
                return method(req, resp, *match.groups())

        raise Error(NOT_FOUND, "No handler for {!r}".format(path))


# Helpers

def find_addresses(host, port=0):
    # In the past we use "" as a special address to bind to all interfaces.
    # Using "" with socket.getaddrinfo() would result into socket.gaierror. To
    # keep backward compatibility with old configurations, replace "" with "0"
    # which acts in the same way and will be bound to IPv4 interface.
    if host == "":
        host = "0"

    for ai in socket.getaddrinfo(host, port, type=socket.SOCK_STREAM):
        # Filter IPv6 link local addresses. Users can use "localhost" or "::1"
        # to listen to local only address.
        # https://en.wikipedia.org/wiki/Link-local_address
        address = ipaddress.ip_address(ai[4][0])
        if address.is_link_local:
            continue

        yield ai



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import logging
import threading

try:
    import yappi
except ImportError:
    yappi = None

from .. import http
from .. import validate

log = logging.getLogger("profile")
lock = threading.Lock()


class Handler:
    """
    Request handler for the /profile/ resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    def post(self, req, resp):
        """
        Start of stop the profiler.
        """
        if yappi is None:
            raise http.Error(http.NOT_FOUND, "yappi is not installed")

        run = validate.enum(req.query, "run", ("y", "n"))
        if run == "y":
            clock = validate.enum(
                req.query, "clock", ("cpu", "wall"), default="cpu")
            self._start_profiling(clock)
        else:
            self._stop_profiling()

    def get(self, req, resp):
        if yappi is None:
            raise http.Error(http.NOT_FOUND, "yappi is not installed")

        msg = {"running": yappi.is_running()}
        resp.send_json(msg)

    def _start_profiling(self, clock):
        with lock:
            if yappi.is_running():
                raise http.Error(
                    http.BAD_REQUEST, "profile is already running")

            log.info("Starting profiling using %r clock", clock)
            yappi.set_clock_type(clock)
            yappi.start(builtins=True, profile_threads=True)

    def _stop_profiling(self):
        with lock:
            if not yappi.is_running():
                raise http.Error(http.BAD_REQUEST, "profile is not running")

            log.info("Stopping profiling, writing profile to %r",
                     self.config.profile.filename)
            yappi.stop()
            stats = yappi.get_func_stats()
            stats.save(self.config.profile.filename, type="pstat")
            yappi.clear_stats()



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pytest

from . import ci
from . import testutil

requires_ipv6 = pytest.mark.skipif(
    not testutil.ipv6_enabled(),
    reason="IPv6 not available")

flaky_in_ovirt_ci = pytest.mark.xfail(
    ci.is_ovirt(), reason="Test is flaky in oVirt CI", strict=False)



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from collections import namedtuple


class ZeroExtent(namedtuple("ZeroExtent", "start,length,zero,hole")):
    """
    An image extent describing image data.

    Fields:
        start (int): offset in bytes.
        length (int): lenth in bytes.
        zero (bool): if True, this area will be read as zeroes.
        hole (bool): if True, this area is unallocated. Reported only for qcow2
            images, meaning that data will be read from the backing chain. Raw
            images do not report holes.

    The extent describes either raw guest data, or raw host data, depending on
    the the backend. For example, file backend always reutrn host data, while
    NBD backend always return guest data.
    """
    __slots__ = ()

    @classmethod
    def from_dict(cls, d):
        """
        Create instance from dict generated by to_dict().
        """
        # Old imageio server did not report holes.
        return cls(d["start"], d["length"], d["zero"], d.get("hole"))

    @property
    def data(self):
        """
        True if this extent may contain non-zero content.
        """
        return not self.zero

    def to_dict(self):
        """
        Crate dict representation.
        """
        return {
            "start": self.start,
            "length": self.length,
            "zero": self.zero,
            "hole": self.hole,
        }


class DirtyExtent(namedtuple("DirtyExtent", "start,length,dirty,zero")):
    """
    An image extent describing dirty areas that have changed since a previous
    checkpoint. This information is available during incremental backup.

    Fields:
        start (int): offset in bytes.
        length (int): lenth in bytes.
        zero (bool): if True, this area will be read as zeroes.
        dirty (bool): if True, this area was changed since the last checkpoint.

    The extent always describes raw guest data.
    """
    __slots__ = ()

    @classmethod
    def from_dict(cls, d):
        """
        Create instance from dict generated by to_dict().
        """
        # Old imageio server did not report zero for dirty extents. If not
        # specified, we assume that the extent is data, and must be copies
        # during incremental backup. If specified, the backup application can
        # optimize the transfer by not copying zero area.
        return cls(d["start"], d["length"], d["dirty"], d.get("zero", False))

    @property
    def data(self):
        """
        True if this extent may contain non-zero content.
        """
        return not self.zero

    def to_dict(self):
        """
        Crate dict representation.
        """
        return {
            "start": self.start,
            "length": self.length,
            "dirty": self.dirty,
            "zero": self.zero,
        }



# ovirt-imageio
# Copyright (C) 2015-2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from distutils.core import setup
from distutils.core import Extension

from ovirt_imageio._internal import version

with open("README.md") as f:
    long_description = f.read()

setup(
    author="oVirt Authors",
    author_email="devel@ovirt.org",
    description="Transfer disk images on oVirt system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GNU GPLv2+",
    name="ovirt-imageio",
    packages=[
        "ovirt_imageio",
        "ovirt_imageio._internal",
        "ovirt_imageio._internal.backends",
        "ovirt_imageio._internal.handlers",
        "ovirt_imageio.client",
        "ovirt_imageio.admin",
    ],
    platforms=["Linux"],
    scripts=["ovirt-imageio", "ovirt-imageioctl"],
    url="https://github.com/oVirt/ovirt-imageio",
    project_urls={
        "Documentation":
            "http://ovirt.github.io/ovirt-imageio/",
        "Bug Tracker":
            "https://bugzilla.redhat.com/enter_bug.cgi?product=ovirt-imageio",
    },
    version=version.string,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Topic :: System :: Archiving :: Backup",
        "Topic :: Utilities",
    ],
    ext_modules=[
        Extension(
            "ovirt_imageio/_internal/ioutil",
            sources=["ovirt_imageio/_internal/ioutil.c"]),
    ]
)



# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pytest

from ovirt_imageio._internal import auth
from ovirt_imageio._internal import config
from ovirt_imageio._internal import errors
from ovirt_imageio._internal import services


@pytest.mark.parametrize("port", [-1, 65536])
def test_invalid_remote_port(port):
    cfg = config.load(["test/conf/daemon.conf"])
    authorizer = auth.Authorizer(cfg)
    cfg.remote.port = port
    with pytest.raises(errors.InvalidConfig):
        services.RemoteService(cfg, authorizer)


@pytest.mark.parametrize("port", [-1, 65536])
def test_invalid_control_port(port):
    cfg = config.load(["test/conf/proxy.conf"])
    authorizer = auth.Authorizer(cfg)
    cfg.control.port = port
    with pytest.raises(errors.InvalidConfig):
        services.ControlService(cfg, authorizer)



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import hashlib
import os
from functools import partial

from . import ioutil

# These settings give best result for Fedora 32 image when using the nbd
# backend. More testing is needed to determine if this is the best default.
BLOCK_SIZE = 4 * 1024**2
ALGORITHM = "blake2b"
DIGEST_SIZE = 32


class Hash:
    """
    Block based hash supporting fast zero block hashing.

    To use this, you must first split the input to block_size legnth blocks.
    Use blkhash.split() to retrun stream of blocks from stream of variable size
    extents.

    Then call update(data) for every data block, and zero(length) for every
    zero block, in the order of the blocks in the file. Zero block hashing is
    optimized by using pre-computed digest instead of hashing zero block.

    If you don't have extents information, split the file to block_size legnth
    blocks and call update(data) in the order of the blcoks. The result will be
    equal but much slower.

    The last block may be shorter if the file is not aligned to block_size.
    """

    def __init__(self, block_size=BLOCK_SIZE, algorithm=ALGORITHM,
                 digest_size=DIGEST_SIZE):
        self._func = getattr(hashlib, algorithm)
        if digest_size:
            self._func = partial(self._func, digest_size=digest_size)
        self._hash = self._func()
        self._block_size = block_size
        self._zero_block_digest = self._func(b"\0" * block_size).digest()

    def update(self, block):
        block_digest = self._func(block).digest()
        self._hash.update(block_digest)

    def zero(self, count):
        if count == self._block_size:
            # Fast path.
            self._hash.update(self._zero_block_digest)
        else:
            # Slow path.
            block_digest = self._func(b"\0" * count).digest()
            self._hash.update(block_digest)

    def digest(self):
        return self._hash.digest()

    def hexdigest(self):
        return self._hash.hexdigest()


def checksum(path, block_size=BLOCK_SIZE, algorithm=ALGORITHM,
             digest_size=DIGEST_SIZE, detect_zeroes=True):
    """
    Compute file checksum without extents information.

    Arguments:
        path (str): Path to image.
        block_size (int): Size of block in bytes. Should be multiple of 4096.
        algorithm (str): One of the algorithms supported py haslib module.
        digest_size (int): Size of hash in bytes, supported only for blake2b
            and blake2s algorithms; specify None for other algorithms.
        detect_zeroes (bool): If True, detect zeroes in the input, speeing up
            the calculation.
    """
    length = os.path.getsize(path)
    block = bytearray(block_size)
    h = Hash(
        block_size=block_size, algorithm=algorithm, digest_size=digest_size)

    with open(path, "rb") as f:
        # Hash full blocks.
        while length >= block_size:
            _read_block(f, block_size, block)
            if detect_zeroes and ioutil.is_zero(block):
                h.zero(block_size)
            else:
                h.update(block)
            length -= block_size

        # Hash last partial block.
        if length:
            with memoryview(block)[:length] as view:
                _read_block(f, length, view)
                h.update(view)

    return {
        "algorithm": algorithm,
        "block_size": block_size,
        "checksum": h.hexdigest(),
    }


def _read_block(f, length, buf):
    pos = 0
    while pos < length:
        with memoryview(buf)[pos:] as view:
            pos += f.readinto(view)


class Block:
    """
    Block descriptor.
    """

    __slots__ = ("start", "length", "zero")

    def __init__(self, start, length, zero):
        self.start = start
        self.length = length
        self.zero = zero

    def merge(self, other, block_size):
        """
        Merge part of another block into this block, possibly converting this
        block to a data block.
        """
        stolen = min(block_size - self.length, other.length)

        # Steal range from other...
        other.start += stolen
        other.length -= stolen

        # And add to myself, possibly converting to data block.
        self.length += stolen
        self.zero &= other.zero

    def split(self, block_size):
        """
        Split another block from this block. Valid only if this block length is
        bigger than block_size.
        """
        assert self.length >= block_size

        block = Block(self.start, block_size, self.zero)
        self.start += block_size
        self.length -= block_size

        return block

    def __repr__(self):
        return (f"Block(start={self.start}, length={self.length}, "
                f"zero={self.zero})")


def split(extents, block_size=BLOCK_SIZE):
    """
    Generate stream of block_size blocks from extents stream.

    Extents smaller than block_size will be merge into a single block of
    block_size length. Merging blocks will convert small zero block to data
    blocks, or steal part of zero block into a data block.

    Extents:  |   data   | zero |  data  |          zero                  |
    Blocks:   |     data     |     data     |     zero     |     zero     |

    Extents larger than block_size are split into multiple block_size length
    blocks.

    Extents:  |             data            |             zero            |
    Blocks:   |     data     |     data     |     zero     |     zero     |

    If the image is not aligned to block_size, the last block length will be
    smaller than block_size.

    Extents:  |     data     |             zero            | data |
    Blocks:   |     data     |     zero     |     zero     | data |
    """
    partial = None

    for extent in extents:
        current = Block(extent.start, extent.length, extent.zero)

        # Try to complete and yield partial block.
        if partial:
            partial.merge(current, block_size)
            if partial.length < block_size:
                continue

            yield partial
            partial = None

        # Yield complete blocks.
        while current.length >= block_size:
            yield current.split(block_size)

        # Keep the partial block for the next extent.
        if current.length:
            partial = current

    if partial:
        yield partial



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
Helpers for handling issues on specific CI environment.
"""

import os


def is_ovirt():
    return "OVIRT_CI" in os.environ


def is_travis():
    return "TRAVIS_CI" in os.environ



# ovirt-imageio
# Copyright (C) 2015-2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from ovirt_imageio._internal import errors


def test_str():
    e = errors.PartialContent(50, 42)
    assert str(e) == "Requested 50 bytes, available 42 bytes"



# ovirt-imageio
# Copyright (C) 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import itertools
import logging
import time

import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import errors
from ovirt_imageio._internal import ops
from ovirt_imageio._internal import util
from ovirt_imageio._internal.auth import Ticket, Authorizer

from test import testutil

CHUNK_SIZE = 8 * 1024**2

log = logging.getLogger("test")


class Context:

    def __init__(self):
        self.closed = False

    def close(self):
        self.closed = True


class Operation:
    """
    Used to fake a ops.Operation object.
    """

    def __init__(self, offset=0, size=0):
        self.offset = offset
        self.size = size
        self.done = 0
        self.canceled = False

    def run(self):
        if self.canceled:
            raise ops.Canceled
        self.done = self.size

    def cancel(self):
        self.canceled = True


@pytest.fixture
def cfg():
    return config.load(["test/conf.d/daemon.conf"])


def test_transfered_nothing(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    assert ticket.transferred() == 0


def test_transfered_inactive_empty_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.run(Operation(0, 0))
    assert ticket.transferred() == 0

    ticket.run(Operation(1000, 0))
    assert ticket.transferred() == 0


def test_transfered_inactive_ordered_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.run(Operation(0, 100))
    assert ticket.transferred() == 100

    ticket.run(Operation(100, 100))
    assert ticket.transferred() == 200

    ticket.run(Operation(200, 100))
    assert ticket.transferred() == 300


def test_transfered_inactive_unordered_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.run(Operation(100, 100))
    assert ticket.transferred() == 100

    ticket.run(Operation(0, 100))
    assert ticket.transferred() == 200

    ticket.run(Operation(200, 100))
    assert ticket.transferred() == 300


def test_transfered_inactive_overlapping_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.run(Operation(0, 120))
    assert ticket.transferred() == 120

    ticket.run(Operation(100, 100))
    assert ticket.transferred() == 200

    ticket.run(Operation(180, 120))
    assert ticket.transferred() == 300


def test_transfered_inactive_non_continuous_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    # Run 2 non-continutes operations
    ticket.run(Operation(0, 100))
    ticket.run(Operation(200, 100))
    assert ticket.transferred() == 200

    # Run last operation filling the hole - with some overlap.
    ticket.run(Operation(80, 120))
    assert ticket.transferred() == 300


def test_transfered_ongoing_concurrent_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    # Start 2 ongoing operations:
    # ongoing: 0-0, 100-100
    # completed:
    op1 = Operation(0, 100)
    ticket._add_operation(op1)
    assert ticket.transferred() == 0
    assert ticket.active()

    op2 = Operation(100, 100)
    ticket._add_operation(op2)
    assert ticket.transferred() == 0
    assert ticket.active()

    # Consume op1 data:
    # ongoing: 0-100, 100-100
    # completed:
    op1.run()
    ticket._remove_operation(op1)
    assert ticket.transferred() == 100
    assert ticket.active()

    # Consume op2 data:
    # ongoing: 0-100, 100-200
    # completed:
    op2.run()
    ticket._remove_operation(op2)
    assert ticket.transferred() == 200
    assert not ticket.active()


def test_transfered_ongoing_overlapping_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    # Start 2 ongoing operations.
    # ongoing: 0-0, 80-80
    # completed:
    op1 = Operation(0, 120)
    op2 = Operation(80, 120)
    ticket._add_operation(op1)
    ticket._add_operation(op2)
    assert ticket.transferred() == 0
    assert ticket.active()

    # Consume op1 data:
    # ongoing: 0-120, 80-80
    # completed:
    op1.run()
    ticket._remove_operation(op1)
    assert ticket.transferred() == 120
    assert ticket.active()

    # Consume op2 data:
    # ongoing: 0-120, 80-200
    # completed:
    op2.run()
    ticket._remove_operation(op2)
    assert ticket.transferred() == 200
    assert not ticket.active()


def test_transfered_ongoing_non_continues_ops(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    # Start 2 ongoing operations.
    # ongoing: 0-0, 200-200
    # completed:
    op1 = Operation(0, 100)
    op2 = Operation(200, 100)
    ticket._add_operation(op1)
    ticket._add_operation(op2)
    assert ticket.transferred() == 0
    assert ticket.active()

    # Consume op1 data:
    # ongoing: 0-100, 200-200
    # completed:
    op1.run()
    ticket._remove_operation(op1)
    assert ticket.transferred() == 100

    # Consume op2 data:
    # ongoing: 0-100, 200-300
    # completed:
    op2.run()
    ticket._remove_operation(op2)
    assert ticket.transferred() == 200


class Client:

    name = "client"
    workers = 1
    io_size = 1024**2

    def run(self):
        start = time.monotonic()
        threads = []
        try:
            for i in range(self.workers):
                t = util.start_thread(self._worker, name=f"{self.name}/{i}")
                threads.append(t)
        finally:
            for t in threads:
                t.join()
        return time.monotonic() - start

    def _worker(self):
        raise NotImplementedError


class Nbdcopy(Client):
    """
    Simulate nbdcopy client calls.

    The image is split to 128 MiB segments, and each worker sends one segment
    at a time to the server.

    Because very worker is sending to a different area, completed ranges are
    always non-continuous, until the segment is completed.

    worker 1:  [=====-      ]
    worker 2:               [=====-      ]
    worker 3:                            [=====-      ]
    worker 4:                                         [=====-      ]

    completed: [=====        =====        =====        =====       ]
    ongoing:   [     -            -            -            -      ]
    """

    name = "nbdcopy"
    segment_size = 128 * 1024**2

    def __init__(self, ticket, workers):
        assert (ticket.size % self.segment_size) == 0
        self.ticket = ticket
        self.image_size = ticket.size
        self.segment_count = self.image_size // self.segment_size
        self.workers = min(workers, self.segment_count)
        self.segment_counter = itertools.count()

    def _worker(self):
        for segment in iter(self._next_segment, None):
            offset = segment * self.segment_size
            log.info("Transferring segment %d offset %d", segment, offset)
            end = offset + self.segment_size
            while offset < end:
                self.ticket.run(Operation(offset, self.io_size))
                offset += self.io_size

    def _next_segment(self):
        n = next(self.segment_counter)
        if n < self.segment_count:
            return n


class Imageio(Client):
    """
    Simulate imageio client calls.

    Workers read extents from a queue and send requests to the server.

    Because all threads are sending to same area, operations are more likely to
    be consecutive, and merged immediately with the previous requests.

    worker 1:  [==      ==      --                                    ]
    worker 2:  [  ==      ==      --                                  ]
    worker 3:  [    ==      ==      --                                ]
    worker 4:  [      ==      ==      --                              ]

    completed: [================                                      ]
    ongoing:   [                --------                              ]
    """

    name = "imageio"

    def __init__(self, ticket, workers):
        assert (ticket.size % self.io_size) == 0
        self.ticket = ticket
        self.image_size = ticket.size
        self.workers = workers
        self.request_count = self.image_size // self.io_size
        self.request_counter = itertools.count()

    def _worker(self):
        for request in iter(self._next_request, None):
            offset = request * self.io_size
            self.ticket.run(Operation(offset, self.io_size))

    def _next_request(self):
        n = next(self.request_counter)
        if n < self.request_count:
            return n


@pytest.mark.benchmark
@pytest.mark.parametrize("workers", [1, 2, 4, 8])
@pytest.mark.parametrize("client_class", [Nbdcopy, Imageio],
                         ids=lambda x: x.name)
@pytest.mark.parametrize("mode", [
    pytest.param(["read"], id="ro"),
    pytest.param(["read", "write"], id="rw"),
])
def test_run_benchmark(cfg, workers, client_class, mode):
    ticket = Ticket(
        testutil.create_ticket(
            size=100 * 1024**3,
            ops=mode),
        cfg)

    client = client_class(ticket, workers)

    elapsed = client.run()

    if mode != ["read", "write"]:
        assert ticket.transferred() == ticket.size

    ops = ticket.size // client.io_size
    print("%d workers, %d ops, %.3f s, %.2f ops/s"
          % (client.workers, ops, elapsed, ops / elapsed))


@pytest.mark.benchmark
@pytest.mark.parametrize("workers", [1, 2, 4, 8])
def test_transferred_benchmark(cfg, workers):
    # Time trransferred call with multiple ongoing and completed operations.
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    ops = 10000

    # Add some completed ranges - assume worst case when ranges are not
    # continues.
    for i in range(workers):
        ticket.run(Operation(i * 1000, 100))

    # Add some ongoing operations - assume worst case when ranges are not
    # continues.
    for i in range(workers):
        ticket._add_operation(Operation(i * 1000 + 200, 100))

    # Time transferred call - merging ongoing and completed ranges.
    start = time.monotonic()
    for i in range(ops):
        ticket.transferred()
    elapsed = time.monotonic() - start

    print("%d workers, %d ops, %.3f s, %.2f ops/s"
          % (workers, ops, elapsed, ops / elapsed))


@pytest.mark.parametrize("arg", [
    "not a dict",
    ["not", "a", "dict"],
    1,
    3.1,
    True,
    False,
    None
])
def test_invalid_argument(arg, cfg):
    with pytest.raises(errors.InvalidTicket):
        Ticket(arg, cfg)


@pytest.mark.parametrize("kw", [
    {"uuid": 1},
    {"size": "not an int"},
    {"ops": "not a list"},
    {"timeout": "not an int"},
    {"url": 1},
    {"transfer_id": 1},
    {"filename": 1},
    {"sparse": 1},
    {"dirty": 1},
    {"inactivity_timeout": "invalid"},
])
def test_invalid_parameter(kw, cfg):
    with pytest.raises(errors.InvalidTicketParameter):
        Ticket(testutil.create_ticket(**kw), cfg)


def test_inactivity_timeout_unset(cfg):
    ticket = Ticket(testutil.create_ticket(inactivity_timeout=None), cfg)
    assert ticket.inactivity_timeout == cfg.daemon.inactivity_timeout


def test_sparse_unset(cfg):
    ticket = Ticket(testutil.create_ticket(), cfg)
    assert not ticket.sparse


def test_sparse(cfg):
    ticket = Ticket(testutil.create_ticket(sparse=True), cfg)
    assert ticket.sparse


def test_dirty_unset(cfg):
    ticket = Ticket(testutil.create_ticket(), cfg)
    assert not ticket.dirty


def test_dirty(cfg):
    ticket = Ticket(testutil.create_ticket(dirty=True), cfg)
    assert ticket.dirty


def test_transfer_id_unset(cfg):
    d = testutil.create_ticket()
    del d["transfer_id"]
    ticket = Ticket(d, cfg)
    assert ticket.transfer_id == f"(ticket/{ticket.uuid[:18]})"


def test_transfer_id(cfg):
    ticket = Ticket(testutil.create_ticket(transfer_id="123"), cfg)
    assert ticket.transfer_id == "123"


def test_repr(cfg):
    ticket = Ticket(
        testutil.create_ticket(
            ops=["read"], filename="tmp_file"),
        cfg)
    ticket_repr = repr(ticket)

    info = ticket.info()
    del info["timeout"]

    for key, value in info.items():
        pair = "%s=%r" % (key, value)
        assert pair in ticket_repr


def test_ticket_run(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    op = Operation(0, 100)
    assert ticket.transferred() == op.done
    assert op.done == 0

    ticket.run(op)

    assert ticket.transferred() == op.done
    assert op.done == 100


def test_cancel_no_connection(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.cancel()

    # Ticket is canceled and can be removed immediately.
    assert ticket.canceled
    info = ticket.info()
    assert info["canceled"]
    assert not info["active"]
    assert info["connections"] == 0


def test_cancel_idle_connection(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ctx = Context()
    ticket.add_context(1, ctx)
    ticket.cancel()

    # Ticket is canceled and can be removed immediately.
    assert ticket.canceled
    assert ctx.closed

    info = ticket.info()
    assert info["canceled"]
    assert not info["active"]

    # The conection context was closed. The connection will be closed when it
    # times out or when a user send the next request.
    assert info["connections"] == 1


def test_cancel_timeout(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    # Add conection - having connections does not block cancelation, but we
    # cannot have ongoing operations without a connection.
    ctx = Context()
    ticket.add_context(1, ctx)

    # Ongoing operation blocks cancel.
    ticket._add_operation(Operation(0, 100))

    # Canceling will time out.
    with pytest.raises(errors.TransferCancelTimeout):
        ticket.cancel(timeout=0.001)

    # Ticket is marked as canceled, but the context was not closed.
    assert ticket.canceled
    assert not ctx.closed

    # Caller can poll ticket "active" property and remove the ticket when the
    # ticket is inactive.
    info = ticket.info()
    assert info["canceled"]
    assert info["active"]
    assert info["connections"] == 1


def test_cancel_async(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ctx = Context()
    ticket.add_context(1, ctx)
    ticket._add_operation(Operation(0, 100))
    ticket.cancel(timeout=0)

    # Ticket is marked as canceled, but the context was not closed.
    assert ticket.canceled
    assert not ctx.closed

    info = ticket.info()
    assert info["canceled"]
    assert info["active"]
    assert info["connections"] == 1


def test_cancel_ongoing_operations(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    # Few connections are using this ticket. Each running an operation.
    ops = []
    for i in range(4):
        ctx = Context()
        op = Operation()
        ticket.add_context(i, ctx)
        ticket._add_operation(op)
        ops.append(op)

    # Add idle connection.
    idle_ctx = Context()
    ticket.add_context(4, idle_ctx)

    ticket.cancel(timeout=0)

    # All ongoing operations are canceled.
    assert all(op.canceled for op in ops)

    # Idle context was not closed.
    assert not idle_ctx.closed


def test_cancel_wait(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    # Add connections using this ticket.
    users = []
    for cid in range(4):
        ctx = Context()
        op = Operation(cid * 1024**2, 1024**2)
        ticket.add_context(cid, ctx)
        ticket._add_operation(op)
        users.append((cid, ctx, op))

    # Add idle connection.
    idle_ctx = Context()
    ticket.add_context(4, idle_ctx)

    def finish_operations():
        time.sleep(0.1)
        for cid, ctx, op in users:
            # Removing operation from a canceled ticket raises, send and error
            # and close the connection.
            try:
                ticket._remove_operation(op)
            except errors.AuthorizationError:
                ticket.remove_context(cid)

    info = ticket.info()
    assert not info["canceled"]
    assert info["connections"] == 5
    assert info["active"]

    t = util.start_thread(finish_operations)
    try:
        ticket.cancel(timeout=10)
    finally:
        t.join()

    info = ticket.info()

    # After the ticket was canceled, ticket is inactive, and all ongoging
    # connnections removed from ticket. The idle connection is left, but its
    # context is closed.

    assert not info["active"]
    assert info["connections"] == 1
    assert all(ctx.closed for cid, ctx, op in users)
    assert idle_ctx.closed


def test_canceled_fail_run_before(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.cancel()

    op = Operation()

    # Running operations must fail.
    with pytest.raises(errors.AuthorizationError):
        ticket.run(op)

    # Operation was not run.
    assert op.done == 0


def test_canceled_fail_run_after(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)

    class Operation:

        def __init__(self):
            self.done = False
            self.canceled = False

        def run(self):
            self.done = True
            ticket.cancel(timeout=0.001)

        def cancel(self):
            self.canceled = True

    op = Operation()

    # If ticket was canceled while ongoing operations are running, ticket run
    # will fail removing the operations.
    with pytest.raises(errors.AuthorizationError):
        ticket.run(op)

    assert op.done


def test_canceled_fail_add_context(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.cancel()

    ctx = Context()

    # Adding new context must fail.
    with pytest.raises(errors.AuthorizationError):
        ticket.add_context(2, ctx)


def test_get_context_missing(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    with pytest.raises(KeyError):
        ticket.get_context(1)


def test_get_context(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ctx = Context()
    ticket.add_context(1, ctx)
    assert ticket.get_context(1) is ctx


def test_remove_context_missing(cfg):
    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ticket.add_context(1, Context())
    assert ticket.info()["connections"] == 1

    ticket.remove_context(2)
    assert ticket.info()["connections"] == 1


def test_remove_context_error(cfg):

    class FailingContext:

        def __init__(self):
            self.count = 1
            self.closed = False

        def close(self):
            if self.count > 0:
                self.count -= 1
                raise RuntimeError("Cannot close yet")
            self.closed = True

    ticket = Ticket(testutil.create_ticket(ops=["read"]), cfg)
    ctx = FailingContext()
    ticket.add_context(1, ctx)

    # If closing a context fails, fail. The ticket cannot be removed
    # until this context is closed successfully.
    with pytest.raises(RuntimeError):
        ticket.cancel(timeout=0)

    assert not ctx.closed

    # Calling again will close context successfully, and the ticket can
    # be removed.
    ticket.cancel(timeout=0)

    assert ctx.closed


def test_authorizer_add(cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    assert ticket.uuid == ticket_info["uuid"]


def test_authorizer_remove_unused(cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    # Ticket is unused so it will be removed.
    auth.remove(ticket_info["uuid"])
    with pytest.raises(KeyError):
        auth.get(ticket_info["uuid"])


def test_authorizer_remove_timeout(cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])

    ctx = Context()
    ticket.add_context(1, ctx)

    idle_ctx = Context()
    ticket.add_context(2, idle_ctx)

    assert ticket.info()["connections"] == 2
    assert not ticket.info()["active"]

    op = Operation(0, 100)
    ticket._add_operation(op)
    assert ticket.info()["active"]

    # Use short timeout to keep the tests fast.
    cfg.control.remove_timeout = 0.001

    # Ticket cannot be removed since it is used by connection 1.
    with pytest.raises(errors.TransferCancelTimeout):
        auth.remove(ticket.uuid)

    # Ticket was not removed.
    assert auth.get(ticket.uuid) is ticket

    # But was canceled.
    info = ticket.info()
    assert info["canceled"]
    assert info["active"]
    assert info["connections"] == 2

    # Contexts not closed.
    assert not ctx.closed
    assert not idle_ctx.closed

    # Ending the operation makes the ticket inactive. The call raise and
    # error handller close the connection, which remove the contenxt
    # from the ticket.
    try:
        ticket._remove_operation(op)
    except errors.AuthorizationError:
        ticket.remove_context(1)

    info = ticket.info()
    assert info["canceled"]
    assert not info["active"]
    assert info["connections"] == 1
    assert ctx.closed

    # Idle context not closed yet.
    assert not idle_ctx.closed

    # Removing the ticket again close the idle context.
    auth.remove(ticket.uuid)
    assert idle_ctx.closed

    # Ticket was removed.
    with pytest.raises(KeyError):
        auth.get(ticket.uuid)


def test_authorizer_remove_async(cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    ctx = Context()
    ticket.add_context(1, ctx)

    idle_ctx = Context()
    ticket.add_context(2, idle_ctx)

    assert not ticket.info()["active"]

    op = Operation(0, 100)
    ticket._add_operation(op)
    assert ticket.info()["active"]

    # Disable the timeout, so removing a ticket cancel the ticket
    # without waiting, and requiring polling the ticket status.
    cfg.control.remove_timeout = 0

    auth.remove(ticket.uuid)

    # Ticket is canceled, but not removed.
    assert ticket.canceled
    assert auth.get(ticket.uuid) is ticket
    info = ticket.info()
    assert info["active"]
    assert info["connections"] == 2
    assert not ctx.closed
    assert not idle_ctx.closed

    # Ending the operation makes the ticket inactive. The call raise and
    # error handller close the connection, which remove the contenxt
    # from the ticket.
    try:
        ticket._remove_operation(op)
    except errors.AuthorizationError:
        ticket.remove_context(1)

    info = ticket.info()
    assert info["canceled"]
    assert not info["active"]
    assert info["connections"] == 1
    assert ctx.closed

    # Idle context not closed yet.
    assert not idle_ctx.closed

    # Removing the ticket again close the idle context.
    auth.remove(ticket.uuid)
    assert idle_ctx.closed

    # Ticket was removed.
    with pytest.raises(KeyError):
        auth.get(ticket.uuid)


def test_authorizer_remove_mising(cfg):
    auth = Authorizer(cfg)
    # Removing missing ticket does not raise.
    auth.remove("no-such-ticket")


def test_authorize_read(cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    assert auth.authorize(ticket.uuid, "read") == ticket

    with pytest.raises(errors.AuthorizationError):
        auth.authorize(ticket.uuid, "write")


def test_authorize_write(cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["write"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    assert auth.authorize(ticket.uuid, "write") == ticket

    # "write" implies also "read".
    assert auth.authorize(ticket.uuid, "read") == ticket


def test_authorizer_no_ticket(cfg):
    auth = Authorizer(cfg)
    with pytest.raises(errors.AuthorizationError):
        auth.authorize("no-such-ticket", "read")


@pytest.mark.parametrize("ops,allowed", [
    (["read"], ["read"]),
    (["write"], ["read", "write"]),
])
def test_authorizer_canceled(ops, allowed, cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=ops)
    auth.add(ticket_info)
    ticket = auth.get(ticket_info["uuid"])

    # Cancelling the ticket disables any operation.
    ticket.cancel()

    for op in allowed:
        with pytest.raises(errors.AuthorizationError):
            auth.authorize(ticket.uuid, op)


def test_authorizer_expired(cfg):
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["write"])
    auth.add(ticket_info)
    ticket = auth.get(ticket_info["uuid"])

    # Extending with zero timeout expire the ticket.
    ticket.extend(0)

    for op in ("read", "write"):
        with pytest.raises(errors.AuthorizationError):
            auth.authorize(ticket.uuid, op)



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
_internal - imageio internal moudules used by the server and client.

Do not use modules form this package; they are internal implemention details
and may change on each release in incompatible way.
"""



# ovirt-imageio
# Copyright (C) 2015-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import ssl
import subprocess


def server_context(certfile, keyfile, cafile=None, enable_tls1_1=False):
    # TODO: Verify client certs
    ctx = ssl.create_default_context(
        purpose=ssl.Purpose.CLIENT_AUTH, cafile=cafile)
    ctx.options |= ssl.OP_NO_TLSv1
    if not enable_tls1_1:
        ctx.options |= ssl.OP_NO_TLSv1_1
    ctx.load_cert_chain(certfile, keyfile=keyfile)
    return ctx


def client_context(cafile=None, enable_tls1_1=False):
    ctx = ssl.create_default_context(
        purpose=ssl.Purpose.SERVER_AUTH, cafile=cafile)
    ctx.options |= ssl.OP_NO_TLSv1
    if not enable_tls1_1:
        ctx.options |= ssl.OP_NO_TLSv1_1
    return ctx


def check_protocol(server_host, server_port, protocol):
    """
    Use openssl command line tool for checking ssl protocol.
    """
    # TODO: Use cert to allow client cert verfication
    cmd = ["openssl",
           "s_client",
           "-connect", "%s:%d" % (server_host, server_port),
           protocol]
    return subprocess.call(cmd)



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import io
import json
import logging

import pytest

from urllib.parse import urlparse

from ovirt_imageio._internal import errors
from ovirt_imageio._internal import extent
from ovirt_imageio._internal import http
from ovirt_imageio._internal import ssl
from ovirt_imageio._internal import uhttp
from ovirt_imageio._internal import util

from ovirt_imageio._internal.backends.http import Backend

log = logging.getLogger("test")


@pytest.fixture(scope="module")
def http_server(tmp_pki):
    server = http.Server(("localhost", 0), http.Connection)
    log.info("Server listening on %r", server.server_address)

    ctx = ssl.server_context(
        tmp_pki.certfile,
        tmp_pki.keyfile,
        cafile=tmp_pki.cafile)
    server.socket = ctx.wrap_socket(server.socket, server_side=True)

    server.url = urlparse(
        "https://localhost:{}/".format(server.server_port))
    server.cafile = tmp_pki.cafile
    server.app = http.Router([])

    t = util.start_thread(
        server.serve_forever,
        kwargs={"poll_interval": 0.1})

    yield server

    server.shutdown()
    t.join()


@pytest.fixture(scope="module")
def uhttp_server(tmp_pki):
    server = uhttp.Server("", uhttp.Connection)
    log.info("Server listening on %r", server.server_address)

    server.app = http.Router([])

    t = util.start_thread(
        server.serve_forever,
        kwargs={"poll_interval": 0.1})

    yield server

    server.shutdown()
    t.join()


# Server personalies.

class Handler:

    def __init__(self, http_server, uhttp_server=None):
        self.features = []
        self.unix_socket = None
        self.dirty = False
        self.requests = 0

        router = http.Router([("/(.*)", self)])
        http_server.app = router
        if uhttp_server:
            uhttp_server.app = router
            self.unix_socket = uhttp_server.server_address

        # Number clusters to make sure get return real data from the server,
        # and put modify server data.
        self.image = bytearray(1024**2)
        for offset in range(0, len(self.image), 64 * 1024):
            self.image[offset] = offset % 256

    def get(self, req, resp, path=None):
        """
        Implement GET with optional Range header.
        """
        self.requests += 1
        if req.range:
            offset = req.range.first
            size = req.range.last + 1 - offset
            if offset + size > len(self.image):
                raise http.Error(
                    http.REQUESTED_RANGE_NOT_SATISFIABLE,
                    "Invalid content range: offset={} size={} length={}"
                    .format(offset, size, len(self.image)))

            resp.status_code = http.PARTIAL_CONTENT
            resp.headers["content-range"] = "bytes {}-{}/{}".format(
                offset, offset + size - 1, len(self.image))
        else:
            offset = 0
            size = len(self.image)
            resp.status_code = http.OK

        resp.headers["content-length"] = size
        self._read(resp, offset, size)

    def put(self, req, resp, path=None):
        """
        Implement PUT with optional Content-Range header.
        """
        self.requests += 1
        offset = req.content_range.first if req.content_range else 0
        size = req.content_length
        self._write(req, offset, size, flush=True)

    def _read(self, resp, offset, size):
        log.debug("READ offset=%s size=%s", offset, size)
        with memoryview(self.image)[offset:offset + size] as view:
            resp.write(view)

    def _write(self, req, offset, size, flush):
        log.debug("WRITE offset=%s size=%s flush=%s", offset, size, flush)
        self.image[offset:offset + size] = req.read()
        self.dirty = not flush


class OldDaemon(Handler):
    """
    Older daemon supported only GET and PUT.
    """


class OldProxy(Handler):
    """
    Old proxy supported GET and PUT, and incompatible version of OPTIONS
    returning non content.
    """

    def options(self, req, resp, ticket_id):
        resp.status_code = http.NO_CONTENT


class Daemon(Handler):
    """
    Modern daemon added OPTIONS exposing capabilities, PATCH/zero, PATCH/flush
    and recently /extents resource.
    """

    def __init__(self, http_server, uhttp_server=None, extents=True):
        super().__init__(http_server, uhttp_server)

        # zero and flush support was introduce with OPTIONS, so we always
        # support both.
        self.features = ["zero", "flush"]
        if extents:
            self.features.append("extents")

        # Extents support was added later. It works only with NBD backend, and
        # emulated otherwise by reporting single non-zero extent.
        # "zero" extents are always available via emulation. "dirty" extents
        # are available only during incremental backup.
        self.extents = {
            "zero": [
                {
                    "start": 0,
                    "length": len(self.image),
                    "zero": True,
                    "hole": False,
                }
            ]
        }

    def options(self, req, resp, path=None):
        """
        Implement OPTIONS.
        """
        self.requests += 1
        log.debug("OPTIONS path=%s", path)
        options = {"features": self.features}
        if self.unix_socket:
            options["unix_socket"] = self.unix_socket
        resp.send_json(options)

    def get(self, req, resp, path=None):
        """
        Override to dispatch "GET /extents" resource.
        """
        if path == "/extents":
            self.requests += 1
            context = req.query.get("context", "zero")
            self._extents(resp, context)
        else:
            super().get(req, resp, path)

    def put(self, req, resp, path=None):
        """
        Implement PUT /?flush=y|n
        """
        self.requests += 1
        offset = req.content_range.first if req.content_range else 0
        size = req.content_length
        flush = req.query.get("flush") == "y"
        self._write(req, offset, size, flush=flush)

    def patch(self, req, resp, path=None):
        """
        Implement PATCH/zero and PATCH/flush.
        """
        self.requests += 1
        msg = json.loads(req.read())
        if msg["op"] == "zero":
            self._zero(msg)
        elif msg["op"] == "flush":
            self._flush()
        else:
            raise http.Error(http.BAD_REQUEST, "Invalid PATCH request")

    def _extents(self, resp, context):
        # Older daemon considered "/extents" as part of the ticket id, and will
        # fail to authorize the request.
        if "extents" not in self.features:
            raise http.Error(http.FORBIDDEN, "No extents for you!")
        if context not in self.extents:
            raise http.Error(http.NOT_FOUND, "No dirty extents for you!")
        log.debug("EXTENTS context=%s", context)
        resp.send_json(self.extents[context])

    def _zero(self, msg):
        offset = msg["offset"]
        size = msg["size"]
        flush = msg["flush"]

        log.debug("ZERO offset=%s size=%s flush=%s", offset, size, flush)

        self.image[offset:offset + size] = b"\0" * size
        self.dirty = not flush

    def _flush(self):
        log.debug("FLUSH")
        self.dirty = False


# Old daemon tests

def test_old_daemon_open(http_server):
    handler = OldDaemon(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        assert b.server_address == http_server.server_address
        assert b.tell() == 0
        assert b.size() == len(handler.image)

        # Client emulates extents (all non-zero).
        assert list(b.extents()) == [
            extent.ZeroExtent(0, b.size(), False, False)
        ]


def test_old_daemon_readinto(http_server):
    handler = OldDaemon(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_readinto(handler, b)


def test_old_daemon_write(http_server):
    handler = OldDaemon(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write(handler, b)
        assert not handler.dirty


def test_old_daemon_zero(http_server):
    handler = OldDaemon(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_zero(handler, b)
        assert not handler.dirty


def test_old_daemon_zero_error(http_server):
    handler = OldDaemon(http_server)

    # With old daemon we emulate zero by putting zeros over the write.

    def fail(req, resp, *args):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.put = fail

    with pytest.raises(http.Error) as e:
        with Backend(http_server.url, http_server.cafile) as b:
            b.zero(4096)

    assert e.value.code == http.FORBIDDEN


def test_old_daemon_read_from(http_server):
    handler = OldDaemon(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_read_from(handler, b)
        assert not handler.dirty


def test_old_daemon_write_to(http_server):
    handler = OldDaemon(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write_to(handler, b)


def test_old_daemon_size_error(http_server):
    handler = OldDaemon(http_server)

    # With old daemon we emulate HEAD with GET. Make it fail to test error
    # handling.

    def fail(req, resp, *args):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.get = fail

    with pytest.raises(http.Error) as e:
        with Backend(http_server.url, None, secure=False) as b:
            b.size()

    assert e.value.code == http.FORBIDDEN


# Old proxy tests

def test_old_proxy_open(http_server):
    handler = OldProxy(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        assert b.server_address == http_server.server_address
        assert b.tell() == 0
        assert b.size() == len(handler.image)

        # Client emulates extents (all non-zero).
        assert list(b.extents()) == [
            extent.ZeroExtent(0, b.size(), False, False)
        ]


def test_old_proxy_readinto(http_server):
    handler = OldProxy(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_readinto(handler, b)


def test_old_proxy_write(http_server):
    handler = OldProxy(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write(handler, b)
        assert not handler.dirty


def test_old_proxy_zero(http_server):
    handler = OldProxy(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_zero(handler, b)
        assert not handler.dirty


def test_old_proxy_read_from(http_server):
    handler = OldProxy(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_read_from(handler, b)
        assert not handler.dirty


def test_old_proxy_write_to(http_server):
    handler = OldProxy(http_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write_to(handler, b)


# Daemon without unix socket tests

def test_daemon_no_unix_socket_open(http_server):
    handler = Daemon(http_server, extents=False)
    with Backend(http_server.url, http_server.cafile) as b:
        assert b.server_address == http_server.server_address
        assert b.tell() == 0
        assert b.size() == len(handler.image)

        # Client emulates extents (all non-zero).
        assert list(b.extents()) == [
            extent.ZeroExtent(0, b.size(), False, False)
        ]


def test_daemon_no_unix_socket_readinto(http_server):
    handler = Daemon(http_server, extents=False)
    with Backend(http_server.url, http_server.cafile) as b:
        check_readinto(handler, b)


def test_daemon_no_unix_socket_write(http_server):
    handler = Daemon(http_server, extents=False)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write(handler, b)
        assert handler.dirty
        b.flush()
        assert not handler.dirty


def test_daemon_no_unix_socket_zero(http_server):
    handler = Daemon(http_server, extents=False)
    with Backend(http_server.url, http_server.cafile) as b:
        check_zero(handler, b)
        assert handler.dirty
        b.flush()
        assert not handler.dirty


def test_daemon_no_unix_socket_read_from(http_server):
    handler = Daemon(http_server, extents=False)
    with Backend(http_server.url, http_server.cafile) as b:
        check_read_from(handler, b)
        assert handler.dirty
        b.flush()
        assert not handler.dirty


def test_daemon_no_unix_socket_write_to(http_server):
    handler = Daemon(http_server, extents=False)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write_to(handler, b)


# Daemon with bad unix socket tests

def test_daemon_bad_unix_socket_open(http_server):
    handler = Daemon(http_server, extents=False)
    handler.unix_socket = "\0bad/socket"
    with Backend(http_server.url, http_server.cafile) as b:
        assert b.server_address == http_server.server_address
        assert b.tell() == 0
        assert b.size() == len(handler.image)

        # Client emulates extents (all non-zero).
        assert list(b.extents()) == [
            extent.ZeroExtent(0, b.size(), False, False)
        ]


# Daemon with no extents support tests

def test_daemon_no_extents_open(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server, extents=False)
    with Backend(http_server.url, http_server.cafile) as b:
        assert b.server_address == uhttp_server.server_address
        assert b.tell() == 0
        assert b.size() == len(handler.image)

        # Client emulates extents (all non-zero).
        assert list(b.extents()) == [
            extent.ZeroExtent(0, b.size(), False, False)
        ]


# Daemon tests

def test_daemon_open(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        assert b.server_address == uhttp_server.server_address
        assert b.tell() == 0
        assert b.size() == len(handler.image)

        # Client reports server extents (all zero).
        assert list(b.extents()) == [
            extent.ZeroExtent(0, b.size(), True, False)
        ]


def test_daemon_close(http_server, uhttp_server):
    Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        pass

    # Closing twice does nothing.
    b.close()

    # But other operations should fail.
    error = "Operation on closed backend"

    buf = bytearray(4096)
    with pytest.raises(ValueError) as e:
        b.write(buf)
    assert str(e.value) == error

    with pytest.raises(ValueError) as e:
        b.readinto(buf)
    assert str(e.value) == error


def test_daemon_open_insecure(http_server, uhttp_server):
    Daemon(http_server, uhttp_server)
    with Backend(http_server.url, None, secure=False) as b:
        assert b.server_address == uhttp_server.server_address


def test_daemon_open_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)

    # Backends is sending OPTIONS request during open.

    def fail(req, resp, *args):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.options = fail

    with pytest.raises(http.Error) as e:
        with Backend(http_server.url, None, secure=False):
            pass

    assert e.value.code == http.FORBIDDEN


def test_daemon_extents_zero(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)

    chunk_size = len(handler.image) // 2
    handler.extents["zero"] = [
        {
            "start": 0,
            "length": chunk_size,
            "zero": False,
            "hole": False,
        },
        {
            "start": chunk_size,
            "length": chunk_size,
            "zero": True,
            "hole": False,
        },
    ]

    with Backend(http_server.url, http_server.cafile) as b:
        # Zero extents are available.
        assert list(b.extents()) == [
            extent.ZeroExtent(0, chunk_size, False, False),
            extent.ZeroExtent(chunk_size, chunk_size, True, False),
        ]

        # Dirty extents are not available.
        with pytest.raises(errors.UnsupportedOperation):
            list(b.extents("dirty"))


def test_daemon_extents_dirty(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)

    chunk_size = len(handler.image) // 2
    handler.extents["dirty"] = [
        {
            "start": 0,
            "length": chunk_size,
            "dirty": True,
            "hole": False,
        },
        {
            "start": chunk_size,
            "length": chunk_size,
            "dirty": False,
            "hole": False,
        },
    ]

    with Backend(http_server.url, http_server.cafile) as b:
        # Both "zero" and "dirty" extents are available.
        assert list(b.extents("zero")) == [
            extent.ZeroExtent(0, b.size(), True, False),
        ]
        assert list(b.extents("dirty")) == [
            extent.DirtyExtent(0, chunk_size, True, False),
            extent.DirtyExtent(chunk_size, chunk_size, False, False),
        ]


def test_daemon_extents_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)

    def fail(req, resp, *args):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.get = fail

    with pytest.raises(http.Error) as e:
        with Backend(http_server.url, None, secure=False) as b:
            list(b.extents())

    assert e.value.code == http.FORBIDDEN


def test_daemon_readinto(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_readinto(handler, b)


def test_daemon_readinto_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_readinto_error(handler, b)


@pytest.mark.parametrize("size", [4096, 42])
def test_daemon_readinto_short(http_server, uhttp_server, size):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        offset = b.size() - size
        buf = bytearray(8192)

        b.seek(offset)
        n = b.readinto(buf)

        assert n == size
        assert b.tell() == b.size()
        assert buf[:size] == handler.image[offset:]
        assert buf[size:] == b"\0" * (8192 - size)


@pytest.mark.parametrize("end_offset", [0, 1])
def test_daemon_readinto_end(http_server, uhttp_server, end_offset):
    _ = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        offset = b.size() + end_offset
        b.seek(offset)
        buf = bytearray(4096)
        n = b.readinto(buf)

        assert n == 0
        assert b.tell() == offset


def test_daemon_write(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write(handler, b)
        assert handler.dirty
        b.flush()
        assert not handler.dirty


def test_daemon_write_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write_error(handler, b)


def test_daemon_zero(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_zero(handler, b)
        assert handler.dirty
        b.flush()
        assert not handler.dirty


def test_daemon_zero_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_zero_error(handler, b)


def test_daemon_flush_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_flush_error(handler, b)


def test_daemon_read_from(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_read_from(handler, b)
        assert handler.dirty
        b.flush()
        assert not handler.dirty


def test_daemon_read_from_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_read_from_error(handler, b)


def test_daemon_write_to(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write_to(handler, b)


def test_daemon_write_to_error(http_server, uhttp_server):
    handler = Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as b:
        check_write_to_error(handler, b)


def test_clone_https(http_server):
    # This server does not provide Unix socket.
    Daemon(http_server)
    with Backend(http_server.url, http_server.cafile) as a:
        # Simulate download flow when we get size from server before cloning
        # the backend.
        a.size()

        with a.clone() as b:
            # Backends are created identical.
            assert a.size() == b.size()
            assert list(a.extents()) == list(b.extents())
            assert a.tell() == b.tell()

            # Modifying one backend does not affect the other.
            assert a.write(b"x" * 4096)
            assert a.tell() == 4096
            assert b.tell() == 0

            # But both show the same content.
            buf = bytearray(4096)
            assert b.readinto(buf)
            assert buf == b"x" * 4096


def test_clone_unix(http_server, uhttp_server):
    Daemon(http_server, uhttp_server)
    with Backend(http_server.url, http_server.cafile) as a:
        # Simulate download flow when we get size from server before cloning
        # the backend.
        a.size()

        with a.clone() as b:
            # Backends are created identical.
            assert a.size() == b.size()
            assert list(a.extents()) == list(b.extents())
            assert a.tell() == b.tell()

            # Modifying one backend does not affect the other.
            assert a.write(b"x" * 4096)
            assert a.tell() == 4096
            assert b.tell() == 0

            # But both show the same content.
            buf = bytearray(4096)
            assert b.readinto(buf)
            assert buf == b"x" * 4096


# Common flows - must works for all variants.

def check_readinto(handler, backend):
    """
    Check single readinto opertion.
    """
    offset = 8192
    length = 4096
    buf = bytearray(length)

    backend.seek(offset)
    backend.readinto(buf)

    assert backend.tell() == offset + length
    assert buf == handler.image[offset:offset + length]


def check_readinto_error(handler, backend):
    """
    Check that remote read error is re-raised in client.
    """
    buf = bytearray(4096)

    def fail(req, resp, tid):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.get = fail

    with pytest.raises(http.Error) as e:
        backend.readinto(buf)

    assert e.value.code == http.FORBIDDEN


def check_write(handler, backend):
    """
    Check single write opertion.
    """
    offset = 8192
    length = 4096
    buf = bytearray(b"x" * length)

    backend.seek(offset)
    backend.write(buf)

    assert backend.tell() == offset + length
    assert buf == handler.image[offset:offset + length]


def check_write_error(handler, backend):
    """
    Check that remote read error is re-raised in client.
    """
    buf = bytearray(4096)

    def fail(req, resp, tid):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.put = fail

    with pytest.raises(http.Error) as e:
        backend.write(buf)

    assert e.value.code == http.FORBIDDEN


def check_zero(handler, backend):
    """
    Check single zero opertion.
    """
    offset = 8192
    length = 512 * 1024

    backend.seek(offset)
    backend.zero(length)

    assert backend.tell() == offset + length
    assert handler.image[offset:offset + length] == b"\0" * length


def check_zero_error(handler, backend):
    """
    Check that remote zero error is re-raised in client.
    """
    def fail(req, resp, tid):
        req.read()
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.patch = fail

    with pytest.raises(http.Error) as e:
        backend.zero(4096)

    assert e.value.code == http.FORBIDDEN


def check_flush_error(handler, backend):
    """
    Check that remote zero error is re-raised in client.
    """
    def fail(req, resp, tid):
        req.read()
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.patch = fail

    with pytest.raises(http.Error) as e:
        backend.flush()

    assert e.value.code == http.FORBIDDEN


def check_read_from(handler, backend):
    """
    Check that we can stream data using small buffer with a single request.
    The length is intentionally not a multiple of the internal buffer size to
    check handling of partial buffers.
    """
    offset = 8192
    length = 600000
    reader = io.BytesIO(b"x" * length)
    buf = bytearray(128 * 1024)
    handler.requests = 0

    backend.seek(offset)
    backend.read_from(reader, length, buf)

    assert handler.requests == 1
    assert backend.tell() == offset + length
    assert handler.image[offset:offset + length] == reader.getvalue()


def check_read_from_error(handler, backend):
    """
    Check remote read error is re-raised in client.
    """
    length = 128 * 1024
    reader = io.BytesIO(b"x" * length)
    buf = bytearray(4096)

    def fail(req, resp, tid):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.put = fail

    with pytest.raises(http.Error) as e:
        backend.read_from(reader, length, buf)

    assert e.value.code == http.FORBIDDEN


def check_write_to(handler, backend):
    """
    Check that we can stream data using small buffer with a single request.
    The length is intentionally not a multiple of the internal buffer size to
    check handling of partial buffers.
    """
    offset = 8192
    length = 600000
    writer = io.BytesIO(b"x" * length)
    buf = bytearray(128 * 1024)
    handler.requests = 0

    backend.seek(offset)
    backend.write_to(writer, length, buf)

    assert handler.requests == 1
    assert backend.tell() == offset + length
    assert writer.getvalue() == handler.image[offset:offset + length]


def check_write_to_error(handler, backend):
    """
    Check that we re-raise remote errors in write_to().
    """
    length = 128 * 1024
    writer = io.BytesIO(b"x" * length)
    buf = bytearray(4096)

    def fail(req, resp, tid):
        raise http.Error(http.FORBIDDEN, "Fake error")

    handler.get = fail

    with pytest.raises(http.Error) as e:
        backend.write_to(writer, length, buf)

    assert e.value.code == http.FORBIDDEN



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
import logging
import socket

log = logging.getLogger("qmp")


class Error(Exception):
    msg = None

    def __str__(self):
        return self.msg.format(self=self)


class CommandFailed(Error):
    msg = "Command {self.cmd} failed: {self.error}"

    def __init__(self, cmd, error):
        self.cmd = cmd
        self.error = error


class NotFound(Error):
    msg = "No node for {self.path}: {self.nodes}"

    def __init__(self, path, nodes):
        self.path = path
        self.nodes = nodes


class Client:

    def __init__(self, sock):
        self.s = socket.socket(socket.AF_UNIX)
        self.s.connect(sock)
        self.r = self.s.makefile("rb")
        self.w = self.s.makefile("wb", 0)
        self._handshake()

    def close(self):
        self.r.close()
        self.w.close()
        self.s.close()

    def execute(self, name, args=None):
        cmd = {"execute": name}
        if args:
            cmd["arguments"] = args
        self._send(cmd)
        while True:
            msg = self._recv()
            if "return" in msg:
                log.debug("Received return: %(return)s", msg)
                return msg["return"]
            elif "error" in msg:
                raise CommandFailed(cmd, msg["error"])
            elif "event" in msg:
                log.debug("Received event: %(event)s: %(data)s" % msg)
            else:
                log.warning("Received unexpected message: %s" % msg)

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        try:
            self.close()
        except Exception:
            if t is not None:
                log.exception("Error closing")

    def _handshake(self):
        msg = self._recv()
        if "QMP" not in msg:
            raise Error("Unexpected response from qemu: %s" % msg)
        log.debug("Talking with qemu %(major)s.%(minor)s.%(micro)s",
                  msg["QMP"]["version"]["qemu"])
        self.execute("qmp_capabilities")

    def _send(self, cmd):
        log.debug("Sending %s", cmd)
        msg = json.dumps(cmd).encode("utf-8")
        self.w.write(msg + b"\n")

    def _recv(self):
        msg = self.r.readline()
        return json.loads(msg.decode("utf-8"))


def find_node(c, path):
    """
    Use connected client c to find a block node using image.

    Raises NotFound if no node is using image.
    """
    r = c.execute("query-block")
    for b in r:
        if "inserted" in b and b["inserted"]["file"] == path:
            return b
    raise NotFound(path, r)



# ovirt-imageio
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import urllib.parse

import pytest


class Backend:
    """
    Wrap a userstorage.Backend, adding a url and context manager interface to
    simplify fixtures.
    """

    def __init__(self, storage, can_detect_sector_size=True):
        if not storage.exists():
            pytest.xfail("Storage {} is not available".format(storage.name))

        self._storage = storage
        self.path = storage.path
        self.url = urllib.parse.urlparse("file:" + storage.path)
        self.sector_size = storage.sector_size
        self.can_detect_sector_size = can_detect_sector_size

    def __enter__(self):
        self._storage.setup()
        return self

    def __exit__(self, *args):
        self._storage.teardown()
